# this will catalog the shot types recorded in the NBA play by play
# not sure how accurate this is it seems to change for the same shots
# I think I have them all added but could be wrong.
# TODO get all the shot types from the hackathon data they sent out and update
# this dictionary
import datetime
import time
import requests


# this dictionary will categorize the event types that happen in the NBA
# play by play
EVENT_TYPE_DICT = {
    1: "shot",
    2: "missed_shot",
    4: "rebound",
    5: "turnover",
    20: "stoppage: out-of-bounds",
    6: "foul",
    3: "free-throw",
    8: "substitution",
    12: "period-start",
    10: "jump-ball",
    9: "team-timeout",
    18: "instant-replay",
    13: "period-end",
    7: "goal-tending",
    0: "game-end",
}


def get_date_games(from_date, to_date):
    """
    Get all the game_ids in a valid date range

    Inputs:
    date_from   - Date to scrape from
    date_to     - Date to scrape to

    Outputs:
    game_ids - List of game_ids in range
    """
    game_ids = []
    from_date = datetime.datetime.strptime(from_date, "%Y-%m-%d")
    to_date = datetime.datetime.strptime(to_date, "%Y-%m-%d")

    # Must check each season in between date range
    for season in range(get_season(from_date), get_season(to_date) + 1):
        url = (
            "http://data.nba.com/data/10s/v2015/json/mobile_teams"
            f"/nba/{season}/league/00_full_schedule.json"
        )
        schedule = requests.get(url).json()
        time.sleep(1)

        for month in schedule["lscd"]:
            if month["mscd"]["g"]:
                # Assume games in order so first in list is first game in month
                cur_month = datetime.datetime.strptime(
                    month["mscd"]["g"][0]["gdte"], "%Y-%m-%d"
                )

                # If first game in month doesn't fall in range no need to check each game for rest of month
                # Convert from_date to beginning of month as that is where cur_month starts
                if (
                    to_date
                    >= cur_month
                    >= (from_date - datetime.timedelta(from_date.day - 1))
                ):
                    for game in month["mscd"]["g"]:
                        # print(game['gdte'])
                        # Check if individual game in date range
                        if (
                            to_date
                            >= datetime.datetime.strptime(game["gdte"], "%Y-%m-%d")
                            >= from_date
                        ):
                            game_ids.append(game["gid"])
    return game_ids


def get_season(date):
    """
    Get Season based on date

    Inputs:
    date  -  time_struct of date

    Outputs:
    season - e.g. 2018
    """
    year = str(date.year)[:4]

    # TODO Refactor this I don't think we need all these if/else statements
    if date > datetime.datetime.strptime("-".join([year, "01-01"]), "%Y-%m-%d"):
        if date < datetime.datetime.strptime("-".join([year, "09-01"]), "%Y-%m-%d"):
            return int(year) - 1
        else:
            return int(year)
    else:
        if date > datetime.datetime.strptime("-".join([year, "07-01"]), "%Y-%m-%d"):
            return int(year)
        else:
            return int(year) - 1



"""
Date: 2019-12-30
Contributor: Matthew Barlowe
twitter: @barloweanalytic
email: matt@barloweanalytics.com

This file contains the main functions to scrape and compile the WNBA api
"""
import sys
import json
import datetime
import requests
import time
import pandas as pd
import numpy as np

from nba_scraper.helper_functions import EVENT_TYPE_DICT, get_season
from nba_scraper.stat_calc_functions import (
    wnba_made_shot,
    wnba_parse_foul,
    wnba_shot_types,
    wnba_seconds_elapsed,
    wnba_points_made,
)

USER_AGENT = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:72.0) Gecko/20100101 Firefox/72.0",
    "Accept": "application/json, text/plain, */*",
    "Accept-Language": "en-US,en;q=0.5",
    "X-NewRelic-ID": "VQECWF5UChAHUlNTBwgBVw==",
    "x-nba-stats-origin": "stats",
    "x-nba-stats-token": "true",
    "Connection": "keep-alive",
    "Referer": "https://stats.nba.com/",
}


def get_player_name(player_id):
    """
    function to get the players name givn a player id

    Inputs:
    player_id   - id of player you want the name of

    Ouputs:
    player_name - full name of player with given player_id
    """
    player_url = f"https://a.data.nba.com/wnba/player/{player_id}"
    player_data = requests.get(player_url, headers=USER_AGENT)
    player_dict = json.loads(player_data.text)
    player_name = (
        f"{player_dict['data']['info']['fn']} {player_dict['data']['info']['ln']}"
    )

    return player_name


def get_team_ids(pbp_df):
    """
    this function gets the home and away team ids

    Inputs:
    pbp_df    - dataframe of the games play by play

    Outputs:
    home_team_id   - team id of the home team
    away_team_id   - team id of the away team
    """

    team_ids = pbp_df["tid"].unique()
    print(team_ids)
    team_ids = [t for t in team_ids if t > 0]
    team_url = f"https://stats.wnba.com/stats/teamdetails?TeamID={team_ids[0]}"
    team_data = requests.get(team_url, headers=USER_AGENT)
    team_data = json.loads(team_data.text)
    print(team_data)

    if (
        pbp_df["home_team_abbrev"].unique()[0]
        == team_data["resultSets"][0]["rowSet"][0][2]
    ):
        home_team_id, away_team_id = team_ids[0], team_ids[1]
    else:
        home_team_id, away_team_id = team_ids[1], team_ids[0]

    return home_team_id, away_team_id


def get_wnba_pbp_api(game_id, quarter, season):
    """
    function gets both JSON requests from the two different APIs if both
    are available and only the stats.nba.com api if not.

    Inputs:
    game_id          - String representing game id
    quarter          - number representing what quarter you want
    season           - number in the format of YYYY representing what the season is

    Outputs:
    wnba_dict         - Dictionary of the JSON response from data.wnba.com api
    """
    wnba_api_url = f"https://data.wnba.com/data/5s/v2015/json/mobile_teams/wnba/{season}/scores/pbp/1{game_id}_{quarter}_pbp.json"

    try:
        wnba_rep = requests.get(wnba_api_url)
    except json.decoder.JSONDecodeError as ex:
        print(ex)
        print(f"This is the stats.nba.com API's output: {wnba_rep.text}")
        sys.exit()

    wnba_dict = wnba_rep.json()

    return wnba_dict


def parse_wnba_pbp(game_id):
    """
    function to parse the JSON output of the api into a dataframe

    Inputs:
    game_id     - Id of game to be parsed
    season      - seaosn game is being played in

    Outputs:
    wnba_pbp_df   - wnba play by play dataframe
    """

    if game_id[2:4] in ["98", "99"]:
        season = f"19{game_id[2:4]}"
    else:
        season = f"20{game_id[2:4]}"
    results = []
    for x in range(1, 15):
        try:
            results.append(get_wnba_pbp_api(game_id, x, season))
        except ValueError:
            break
    dfs = []
    for period, v2_dict in enumerate(results):
        pbp_v2_df = pd.DataFrame(v2_dict["g"]["pla"])
        pbp_v2_df.columns = list(map(str.lower, pbp_v2_df.columns))
        pbp_v2_df["period"] = period + 1
        dfs.append(pbp_v2_df)

    pbp_df = pd.concat(dfs)

    pbp_df["game_date"] = results[0]["g"]["gcode"].split("/")[0]
    pbp_df["game_date"] = pd.to_datetime(pbp_df["game_date"], format="%Y%m%d")
    pbp_df["away_team_abbrev"] = results[0]["g"]["gcode"].split("/")[1][:3]
    pbp_df["home_team_abbrev"] = results[0]["g"]["gcode"].split("/")[1][3:]
    pbp_df["seconds_elapsed"] = pbp_df.apply(wnba_seconds_elapsed, axis=1)
    pbp_df["shot_type"] = pbp_df.apply(wnba_shot_types, axis=1)
    pbp_df["game_id"] = game_id

    # create column whether shot was succesful or not
    pbp_df["shot_made"] = pbp_df.apply(wnba_made_shot, axis=1)

    # calculate event length of each event in seconds
    pbp_df["event_length"] = pbp_df["seconds_elapsed"] - pbp_df[
        "seconds_elapsed"
    ].shift(1)

    # determine whether shot was a three pointer
    pbp_df["is_three"] = np.where(pbp_df["de"].str.contains("3pt").fillna(False), 1, 0,)

    pbp_df["event_type_de"] = pbp_df[["etype"]].replace({"etype": EVENT_TYPE_DICT})

    # create a column that says whether the shot was blocked or not
    pbp_df["is_block"] = np.where(pbp_df["de"].str.contains("BLK"), 1, 0,)

    # determine points earned
    pbp_df["points_made"] = pbp_df.apply(wnba_points_made, axis=1)

    # create columns that determine if rebound is offenseive or deffensive
    pbp_df["is_o_rebound"] = np.where(
        (pbp_df["event_type_de"] == "rebound")
        & (pbp_df["tid"] == pbp_df["tid"].shift(1))
        & (pbp_df["pid"] != 0),
        1,
        0,
    )
    pbp_df["is_d_rebound"] = np.where(
        (pbp_df["event_type_de"] == "rebound")
        & (pbp_df["tid"] != pbp_df["tid"].shift(1))
        & (pbp_df["pid"] != 0),
        1,
        0,
    )

    # create columns to determine turnovers and steals
    pbp_df["is_turnover"] = np.where(
        pbp_df["de"].str.contains("Turnover").fillna(False), 1, 0,
    )
    pbp_df["is_steal"] = np.where(
        pbp_df["de"].str.contains("Steal").fillna(False), 1, 0,
    )

    # determine what type of fouls are being commited
    pbp_df["foul_type"] = pbp_df.apply(wnba_parse_foul, axis=1)

    # determine if a shot is a putback off an offensive reboundk
    pbp_df["is_putback"] = np.where(
        (pbp_df["is_o_rebound"].shift(1) == 1) & (pbp_df["event_length"] <= 3), 1, 0
    )
    pbp_df["home_team_id"], pbp_df["away_team_id"] = get_team_ids(pbp_df)

    return pbp_df


def get_wnba_lineup(game_id, period):
    """
    function pulls the possible lineups for the given period and game id for
    both the away and home teams

    Inputs:
    game_id            - id of game
    period             - period of game

    Outputs:
    lineup_req_dict    - dictionary of lineup api request respons
    """

    if period <= 4:
        start_range = (((period - 1) * 600) * 10) + 5
    else:
        start_range = ((((period - 5) * 300) + 2880) * 10) + 5

    end_range = start_range + 2000

    url = (
        f"https://stats.wnba.com/stats/boxscoreadvancedv2/?gameId=1{game_id}&"
        f"startPeriod={period}&endPeriod={period}&startRange={start_range}&"
        f"endRange={end_range}&rangeType=2"
    )

    lineups_req = requests.get(url, headers=USER_AGENT)
    lineup_req_dict = json.loads(lineups_req.text)

    return lineup_req_dict


def get_lineup(period_df, lineups, dataframe):
    """
    this function calculates the lineups for each team at each event and then
    appends it to the current dataframe. This only works for one period at a
    time

    Inputs:
    period_df         - the main game pbp dataframe subsetted to only one period
                        in the game
    lineups           - lineup api response dictionary
    dataframe         - full game dataframe. This is passed to get players name from
                        id in case the player didn't have an event in that period.

    Outputs:
    lineup_df     - period_df with each teams lineups calculate and added to the
                    dataframe
    """

    home_team = period_df["home_team_id"].unique()[0]
    away_team = period_df["away_team_id"].unique()[0]
    players = lineups["resultSets"][0]["rowSet"]
    home_ids_names = [(p[4], p[5]) for p in players if p[1] == home_team]
    away_ids_names = [(p[4], p[5]) for p in players if p[1] == away_team]

    # gets the index of the first sub for home and away to get the players who started
    # the period by subsetting the dataframe to all actions before the first sub for
    # each team
    # pulls the unique values from the whole period dataframe if there are no subs
    # then it just pulls the unique ids from the from the dataframe itself because
    # the away/home indexes will be an empty list

    if len(away_ids_names) != 5:

        subs_df = period_df[(period_df.event_type_de == "substitution")]
        away_subs = subs_df[subs_df["tid"] == subs_df["away_team_id"]]
        away_indexes = list(away_subs.index)

        try:
            away_starting_line = list(
                period_df[
                    (period_df["tid"] == period_df["away_team_id"].unique()[0])
                    & (~pd.isnull(period_df["pid"]))
                    & (period_df.is_block == 0)
                    & (period_df.is_steal == 0)
                ]
                .loc[: away_indexes[0], :]["pid"]
                .unique()
            )
        except IndexError:
            away_starting_line = list(
                period_df[
                    (period_df["tid"] == period_df["away_team_id"].unique()[0])
                    & (~pd.isnull(period_df["pid"]))
                    & (period_df.is_block == 0)
                    & (period_df.is_steal == 0)
                ]["pid"].unique()
            )

        if away_starting_line != 5:
            starting_lineup = set()
            subs = set()
            for i in range(period_df.shape[0]):
                if (
                    period_df.iloc[i, :]["tid"] == period_df["away_team_id"].unique()[0]
                    and period_df.iloc[i, :]["is_block"] == 0
                    and period_df.iloc[i, :]["is_steal"] == 0
                ):
                    if period_df.iloc[i, :]["event_type_de"] != "substitution":
                        if (
                            period_df.iloc[i, :]["pid"] != 0
                            and period_df.iloc[i, :]["pid"] not in subs
                        ):
                            starting_lineup.add(period_df.iloc[i, :]["pid"])
                    else:
                        if period_df.iloc[i, :]["epid"] not in starting_lineup:
                            subs.add(period_df.iloc[i, :]["epid"])
                        if period_df.iloc[i, :]["pid"] not in subs:
                            starting_lineup.add(period_df.iloc[i, :]["pid"])

                    if len(starting_lineup) == 5:
                        break
            if len(away_ids_names) < 5:
                away_ids_names = [(x, get_player_name(x)) for x in starting_lineup]
            elif len(home_ids_names) > 5:
                away_ids_names = [
                    (p[0], p[1]) for p in away_ids_names if p[0] not in subs
                ]

    if len(home_ids_names) != 5:
        # subsets main dataframe by period and subsets into a home and away subs
        subs_df = period_df[(period_df.event_type_de == "substitution")]
        home_subs = subs_df[subs_df["tid"] == subs_df["home_team_id"]]
        home_indexes = list(home_subs.index)

        try:
            home_starting_line = list(
                period_df[
                    (period_df["tid"] == period_df["home_team_id"].unique()[0])
                    & (~pd.isnull(period_df["pid"]))
                    & (period_df.is_block == 0)
                    & (period_df.is_steal == 0)
                ]
                .loc[: home_indexes[0], :]["pid"]
                .unique()
            )
        except IndexError:
            home_starting_line = list(
                period_df[
                    (period_df["tid"] == period_df["home_team_id"].unique()[0])
                    & (~pd.isnull(period_df["pid"]))
                    & (period_df.is_block == 0)
                    & (period_df.is_steal == 0)
                ]["pid"].unique()
            )

        if home_starting_line != 5:
            starting_lineup = set()
            subs = set()
            for i in range(period_df.shape[0]):
                if (
                    period_df.iloc[i, :]["tid"] == period_df["home_team_id"].unique()[0]
                    and period_df.iloc[i, :]["is_block"] == 0
                    and period_df.iloc[i, :]["is_steal"] == 0
                ):
                    if period_df.iloc[i, :]["event_type_de"] != "substitution":
                        if (
                            period_df.iloc[i, :]["pid"] != 0
                            and period_df.iloc[i, :]["pid"] not in subs
                        ):
                            starting_lineup.add(period_df.iloc[i, :]["pid"])
                    else:
                        if period_df.iloc[i, :]["epid"] not in starting_lineup:
                            subs.add(period_df.iloc[i, :]["epid"])
                        if period_df.iloc[i, :]["pid"] not in subs:
                            starting_lineup.add(period_df.iloc[i, :]["pid"])

                    if len(starting_lineup) == 5:
                        break
            if len(home_ids_names) < 5:
                home_ids_names = [(x, get_player_name(x)) for x in starting_lineup]
            elif len(home_ids_names) > 5:
                home_ids_names = [
                    (p[0], p[1]) for p in home_ids_names if p[0] not in subs
                ]

    if period_df.shape[1] == 37:
        period_df.loc[:, "dummy_col"] = ""
    # creating columns to populate with players on the court
    period_df.loc[:, "home_player_1"] = ""
    period_df.loc[:, "home_player_1_id"] = ""
    period_df.loc[:, "home_player_2"] = ""
    period_df.loc[:, "home_player_2_id"] = ""
    period_df.loc[:, "home_player_3"] = ""
    period_df.loc[:, "home_player_3_id"] = ""
    period_df.loc[:, "home_player_4"] = ""
    period_df.loc[:, "home_player_4_id"] = ""
    period_df.loc[:, "home_player_5"] = ""
    period_df.loc[:, "home_player_5_id"] = ""
    period_df.loc[:, "away_player_1"] = ""
    period_df.loc[:, "away_player_1_id"] = ""
    period_df.loc[:, "away_player_2"] = ""
    period_df.loc[:, "away_player_2_id"] = ""
    period_df.loc[:, "away_player_3"] = ""
    period_df.loc[:, "away_player_3_id"] = ""
    period_df.loc[:, "away_player_4"] = ""
    period_df.loc[:, "away_player_4_id"] = ""
    period_df.loc[:, "away_player_5"] = ""
    period_df.loc[:, "away_player_5_id"] = ""
    # add players to the columns by looping through the dataframe and putting the
    # players in for each row using the starting lineup list. If there is a
    # substitution event then the player coming on replaces the player going off in
    # the list this is done for the whole period

    for i in range(period_df.shape[0]):
        if (
            period_df.iloc[i, :]["event_type_de"] == "substitution"
            and period_df.iloc[i, :]["tid"] == home_team
        ):
            home_ids_names = [
                ids for ids in home_ids_names if ids[0] != period_df.iloc[i, :]["pid"]
            ]
            player_url = (
                f"https://a.data.nba.com/wnba/player/{period_df.iloc[i, :]['epid']}"
            )
            player_data = requests.get(player_url, headers=USER_AGENT)
            player_dict = json.loads(player_data.text)
            player_name = f"{player_dict['data']['info']['fn']} {player_dict['data']['info']['ln']}"
            home_ids_names.append((period_df.iloc[i, :]["epid"], player_name))
            period_df.iat[i, 39] = home_ids_names[0][0]
            period_df.iat[i, 38] = home_ids_names[0][1]
            period_df.iat[i, 41] = home_ids_names[1][0]
            period_df.iat[i, 40] = home_ids_names[1][1]
            period_df.iat[i, 43] = home_ids_names[2][0]
            period_df.iat[i, 42] = home_ids_names[2][1]
            period_df.iat[i, 45] = home_ids_names[3][0]
            period_df.iat[i, 44] = home_ids_names[3][1]
            period_df.iat[i, 47] = home_ids_names[4][0]
            period_df.iat[i, 46] = home_ids_names[4][1]
            period_df.iat[i, 49] = away_ids_names[0][0]
            period_df.iat[i, 48] = away_ids_names[0][1]
            period_df.iat[i, 51] = away_ids_names[1][0]
            period_df.iat[i, 50] = away_ids_names[1][1]
            period_df.iat[i, 53] = away_ids_names[2][0]
            period_df.iat[i, 52] = away_ids_names[2][1]
            period_df.iat[i, 55] = away_ids_names[3][0]
            period_df.iat[i, 54] = away_ids_names[3][1]
            period_df.iat[i, 57] = away_ids_names[4][0]
            period_df.iat[i, 56] = away_ids_names[4][1]
        elif (
            period_df.iloc[i, :]["event_type_de"] == "substitution"
            and period_df.iloc[i, :]["tid"] == away_team
        ):
            away_ids_names = [
                ids for ids in away_ids_names if ids[0] != period_df.iloc[i, :]["pid"]
            ]
            player_url = (
                f"https://a.data.nba.com/wnba/player/{period_df.iloc[i, :]['epid']}"
            )
            player_data = requests.get(player_url, headers=USER_AGENT)
            player_dict = json.loads(player_data.text)
            player_name = f"{player_dict['data']['info']['fn']} {player_dict['data']['info']['ln']}"
            away_ids_names.append((period_df.iloc[i, :]["epid"], player_name))
            period_df.iat[i, 39] = home_ids_names[0][0]
            period_df.iat[i, 38] = home_ids_names[0][1]
            period_df.iat[i, 41] = home_ids_names[1][0]
            period_df.iat[i, 40] = home_ids_names[1][1]
            period_df.iat[i, 43] = home_ids_names[2][0]
            period_df.iat[i, 42] = home_ids_names[2][1]
            period_df.iat[i, 45] = home_ids_names[3][0]
            period_df.iat[i, 44] = home_ids_names[3][1]
            period_df.iat[i, 47] = home_ids_names[4][0]
            period_df.iat[i, 46] = home_ids_names[4][1]
            period_df.iat[i, 49] = away_ids_names[0][0]
            period_df.iat[i, 48] = away_ids_names[0][1]
            period_df.iat[i, 51] = away_ids_names[1][0]
            period_df.iat[i, 50] = away_ids_names[1][1]
            period_df.iat[i, 53] = away_ids_names[2][0]
            period_df.iat[i, 52] = away_ids_names[2][1]
            period_df.iat[i, 55] = away_ids_names[3][0]
            period_df.iat[i, 54] = away_ids_names[3][1]
            period_df.iat[i, 57] = away_ids_names[4][0]
            period_df.iat[i, 56] = away_ids_names[4][1]
        else:
            period_df.iat[i, 39] = home_ids_names[0][0]
            period_df.iat[i, 38] = home_ids_names[0][1]
            period_df.iat[i, 41] = home_ids_names[1][0]
            period_df.iat[i, 40] = home_ids_names[1][1]
            period_df.iat[i, 43] = home_ids_names[2][0]
            period_df.iat[i, 42] = home_ids_names[2][1]
            period_df.iat[i, 45] = home_ids_names[3][0]
            period_df.iat[i, 44] = home_ids_names[3][1]
            period_df.iat[i, 47] = home_ids_names[4][0]
            period_df.iat[i, 46] = home_ids_names[4][1]
            period_df.iat[i, 49] = away_ids_names[0][0]
            period_df.iat[i, 48] = away_ids_names[0][1]
            period_df.iat[i, 51] = away_ids_names[1][0]
            period_df.iat[i, 50] = away_ids_names[1][1]
            period_df.iat[i, 53] = away_ids_names[2][0]
            period_df.iat[i, 52] = away_ids_names[2][1]
            period_df.iat[i, 55] = away_ids_names[3][0]
            period_df.iat[i, 54] = away_ids_names[3][1]
            period_df.iat[i, 57] = away_ids_names[4][0]
            period_df.iat[i, 56] = away_ids_names[4][1]

    return period_df


def wnba_main_scrape(game_id):
    """
    This is the main function which ties everything together and will be imported
    into nba_scraper module as the hook to scrape nba games

    Inputs:
    game_id      - WNBA game id to be scraped

    Outputs:
    game_df      - WNBA dataframe of the play by play

    WNBA has 204 games in a season with the same game_id structure as nba
    they add a 1 to the front of the game_id for some reason
    """

    pbp_df = parse_wnba_pbp(game_id)

    periods = []

    for period in range(1, pbp_df["period"].max() + 1):
        lineups = get_wnba_lineup(game_id, period)
        periods.append(
            get_lineup(pbp_df[pbp_df["period"] == period].copy(), lineups, pbp_df,)
        )

    pbp_df = pd.concat(periods)

    return pbp_df


def main():
    # parse_wnba_pbp("021900122")
    wnba_main_scrape("021900001")


if __name__ == "__main__":
    main()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()
setup(
    name="nba_scraper",
    packages=["nba_scraper"],
    version="1.0.10",
    license="GNU General Public License v3.0",
    description="A Python package to scrape the NBA api and return a play by play file",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Matthew Barlowe",
    author_email="matt@barloweanalytics.com",
    url="https://github.com/mcbarlowe/nba_scraper",
    download_url="https://github.com/mcbarlowe/nba_scraper/archive/v1.0.10.tar.gz",
    keywords=["basketball", "NBA", "scraper"],
    install_requires=["requests", "pandas", "numpy"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.7",
    ],
)



"""
Date: 2019-01-02
Contributor: Matthew Barlowe
Twitter: @barloweanalytic
Email: matt@barloweanalytics.com

This file contains the main functions to scrape and compile the NBA api and
return a CSV file of the pbp for the provided game
"""
import sys
import json
import datetime
import requests
import time
import pandas as pd
import numpy as np

# TODO probably need to fix these to import modularly correctly
from nba_scraper.helper_functions import EVENT_TYPE_DICT, get_season
from nba_scraper.stat_calc_functions import (
    made_shot,
    parse_foul,
    parse_shot_types,
    create_seconds_elapsed,
    calc_points_made,
)

# TODO look at replacing this with the fake-useragent package Matt Barlowe 2019-12-04
# have to pass this to the requests function or the api will return a 403 code
USER_AGENT = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:72.0) Gecko/20100101 Firefox/72.0",
    "Accept": "application/json, text/plain, */*",
    "Accept-Language": "en-US,en;q=0.5",
    "X-NewRelic-ID": "VQECWF5UChAHUlNTBwgBVw==",
    "x-nba-stats-origin": "stats",
    "x-nba-stats-token": "true",
    "Connection": "keep-alive",
    "Referer": "https://stats.nba.com/",
}


def get_date_games(from_date, to_date):
    """
    Get all the game_ids in a valid date range
    Inputs:
    date_from   - Date to scrape from
    date_to     - Date to scrape to
    Outputs:
    game_ids - List of game_ids in range
    """
    game_ids = []
    from_date = datetime.datetime.strptime(from_date, "%Y-%m-%d")
    to_date = datetime.datetime.strptime(to_date, "%Y-%m-%d")

    # Must check each season in between date range
    for season in range(get_season(from_date), get_season(to_date) + 1):
        url = (
            "http://data.nba.com/data/10s/v2015/json/mobile_teams"
            f"/nba/{season}/league/00_full_schedule.json"
        )
        schedule = requests.get(url).json()
        time.sleep(1)

        for month in schedule["lscd"]:
            if month["mscd"]["g"]:
                # Assume games in order so first in list is first game in month
                cur_month = datetime.datetime.strptime(
                    month["mscd"]["g"][0]["gdte"], "%Y-%m-%d"
                )

                # If first game in month doesn't fall in range no need to check each game for rest of month
                # Convert from_date to beginning of month as that is where cur_month starts
                if (
                    to_date
                    >= cur_month
                    >= (from_date - datetime.timedelta(from_date.day - 1))
                ):
                    for game in month["mscd"]["g"]:
                        # print(game['gdte'])
                        # Check if individual game in date range
                        if (
                            to_date
                            >= datetime.datetime.strptime(game["gdte"], "%Y-%m-%d")
                            >= from_date
                        ):
                            game_ids.append(game["gid"])
    return game_ids


def scrape_pbp(v2_dict):
    """
    This function scrapes both of the pbp urls and returns a joined/cleaned
    pbp dataframe

    Inputs:
    v2_dict     - stats.nba.com api response

    Outputs:
    clean_df - final cleaned dataframe
    """

    # converting stats.nba.com json into pandas dataframe
    pbp_v2_df = pd.DataFrame(
        v2_dict["resultSets"][0]["rowSet"], columns=v2_dict["resultSets"][0]["headers"]
    )
    pbp_v2_df.columns = list(map(str.lower, pbp_v2_df.columns))

    # pulling the home and away team abbreviations and the game date

    # TODO look at these and see if they are jump ball violations which is
    # TODO eventmsgtype 7 instead of 10
    if pbp_v2_df.game_id.unique()[0] == "0020200577":
        home_team_abbrev = "WAS"
        away_team_abbrev = "DEN"
        pbp_v2_df["home_team_abbrev"] = "WAS"
        pbp_v2_df["away_team_abbrev"] = "DEN"

    elif pbp_v2_df.game_id.unique()[0] == "0020300286":
        home_team_abbrev = "NOH"
        away_team_abbrev = "MIL"
        pbp_v2_df["home_team_abbrev"] = "NOH"
        pbp_v2_df["away_team_abbrev"] = "MIL"

    elif pbp_v2_df.game_id.unique()[0] == "0021700259":
        home_team_abbrev = "HOU"
        away_team_abbrev = "DEN"
        pbp_v2_df["home_team_abbrev"] = "HOU"
        pbp_v2_df["away_team_abbrev"] = "DEN"
    elif pbp_v2_df.game_id.unique()[0] == "0021700477":
        home_team_abbrev = "OKC"
        away_team_abbrev = "ATL"
        pbp_v2_df["home_team_abbrev"] = "OKC"
        pbp_v2_df["away_team_abbrev"] = "ATL"
    elif pbp_v2_df.game_id.unique()[0] == "0021700957":
        home_team_abbrev = "CHA"
        away_team_abbrev = "PHI"
        pbp_v2_df["home_team_abbrev"] = "CHA"
        pbp_v2_df["away_team_abbrev"] = "PHI"
    elif pbp_v2_df.game_id.unique()[0] == "0021900251":
        home_team_abbrev = "DAL"
        away_team_abbrev = "LAC"
        pbp_v2_df["home_team_abbrev"] = "DAL"
        pbp_v2_df["away_team_abbrev"] = "LAC"
    elif pbp_v2_df.game_id.unique()[0] == "0021900539":
        home_team_abbrev = "CHA"
        away_team_abbrev = "IND"
        pbp_v2_df["home_team_abbrev"] = "CHA"
        pbp_v2_df["away_team_abbrev"] = "IND"
    else:
        if (
            pd.isnull(
                pbp_v2_df[pbp_v2_df["eventmsgtype"] == 10][["homedescription"]].values[
                    0
                ]
            )
            == 1
        ):
            home_team_abbrev = pbp_v2_df[pbp_v2_df["eventmsgtype"] == 10][
                "player2_team_abbreviation"
            ].iloc[0]
            away_team_abbrev = pbp_v2_df[pbp_v2_df["eventmsgtype"] == 10][
                "player1_team_abbreviation"
            ].iloc[0]
        else:
            home_team_abbrev = pbp_v2_df[pbp_v2_df["eventmsgtype"] == 10][
                "player1_team_abbreviation"
            ].iloc[0]
            away_team_abbrev = pbp_v2_df[pbp_v2_df["eventmsgtype"] == 10][
                "player2_team_abbreviation"
            ].iloc[0]

        pbp_v2_df["home_team_abbrev"] = home_team_abbrev
        pbp_v2_df["away_team_abbrev"] = away_team_abbrev

    clean_df = pbp_v2_df

    print(
        clean_df[clean_df["player1_team_abbreviation"] == home_team_abbrev][
            "player1_team_id"
        ]
        .astype(int)
        .unique()
    )
    # code to properly get the team ids as the scientific notation cuts off some digits
    clean_df.loc[:, "home_team_id"] = (
        clean_df[clean_df["player1_team_abbreviation"] == home_team_abbrev][
            "player1_team_id"
        ]
        .astype(int)
        .unique()[0]
    )

    clean_df.loc[:, "away_team_id"] = (
        clean_df[clean_df["player1_team_abbreviation"] == away_team_abbrev][
            "player1_team_id"
        ]
        .astype(int)
        .unique()[0]
    )

    clean_df["game_date"] = ""

    if clean_df["game_id"].unique()[0][3:5] == "99":
        clean_df["season"] = 2000
    elif clean_df["game_id"].unique()[0][3:5] == "00":
        clean_df["season"] = 2001
    else:
        clean_df.loc[
            :, ("season")
        ] = f"20{int(clean_df['game_id'].unique()[0][3:5])+1:02}"
    # TODO columns to pull out [['evt', 'locX', 'locY', 'hs', 'vs', 'de']]
    # create an event team colum
    clean_df["event_team"] = np.where(
        clean_df["homedescription"].isnull(),
        clean_df["away_team_abbrev"],
        np.where(
            clean_df["visitordescription"].isnull(),
            clean_df["home_team_abbrev"],
            np.where(
                (clean_df["homedescription"].str.contains("Turnover"))
                | (clean_df["homedescription"].str.contains("MISS")),
                clean_df["home_team_abbrev"],
                clean_df["away_team_abbrev"],
            ),
        ),
    )

    # create and event type description column
    clean_df["event_type_de"] = clean_df[["eventmsgtype"]].replace(
        {"eventmsgtype": EVENT_TYPE_DICT}
    )

    # DON'T DELETE THIS WILL BRAKE WHOLE PROGRAM
    clean_df["shot_type_de"] = ""

    # create column whether shot was succesful or not
    clean_df["shot_made"] = clean_df.apply(made_shot, axis=1)

    # create a column that says whether the shot was blocked or not
    clean_df["is_block"] = np.where(
        clean_df["homedescription"].str.contains("BLOCK")
        | clean_df["visitordescription"].str.contains("BLOCK"),
        1,
        0,
    )
    # parse mtype column to get all the shot types being taken
    clean_df["shot_type"] = clean_df.apply(parse_shot_types, axis=1)

    # Clean time to get a seconds elapsed column
    clean_df["seconds_elapsed"] = clean_df.apply(create_seconds_elapsed, axis=1)

    # calculate event length of each event in seconds
    clean_df["event_length"] = clean_df["seconds_elapsed"] - clean_df[
        "seconds_elapsed"
    ].shift(1)

    # determine whether shot was a three pointer
    clean_df["is_three"] = np.where(
        (clean_df["homedescription"].str.contains("3PT")).fillna(False)
        | (clean_df["visitordescription"].str.contains("3PT")).fillna(False),
        1,
        0,
    )

    # determine points earned
    clean_df["points_made"] = clean_df.apply(calc_points_made, axis=1)

    # create columns that determine if rebound is offenseive or deffensive
    clean_df["is_o_rebound"] = np.where(
        (clean_df["event_type_de"] == "rebound")
        & (clean_df["event_team"] == clean_df["event_team"].shift(1))
        & (
            ~clean_df["player1_id"].isin(
                [clean_df.home_team_id.unique()[0], clean_df.away_team_id.unique()[0]]
            )
        ),
        1,
        0,
    )
    clean_df["is_d_rebound"] = np.where(
        (clean_df["event_type_de"] == "rebound")
        & (clean_df["event_team"] != clean_df["event_team"].shift(1))
        & (
            ~clean_df["player1_id"].isin(
                [clean_df.home_team_id.unique()[0], clean_df.away_team_id.unique()[0]]
            )
        ),
        1,
        0,
    )

    # create columns to determine turnovers and steals
    clean_df["is_turnover"] = np.where(
        (clean_df["homedescription"].str.contains("Turnover")).fillna(False)
        | (clean_df["visitordescription"].str.contains("Turnover")).fillna(False),
        1,
        0,
    )
    clean_df["is_steal"] = np.where(
        (clean_df["homedescription"].str.contains("STEAL")).fillna(False)
        | (clean_df["visitordescription"].str.contains("STEAL")).fillna(False),
        1,
        0,
    )

    # determine what type of fouls are being commited
    clean_df["foul_type"] = clean_df.apply(parse_foul, axis=1)

    # determine if a shot is a putback off an offensive reboundk
    clean_df["is_putback"] = np.where(
        (clean_df["is_o_rebound"].shift(1) == 1) & (clean_df["event_length"] <= 3), 1, 0
    )

    return clean_df


def get_pbp_api(game_id):
    """
    function gets both JSON requests from the two different APIs if both
    are available and only the stats.nba.com api if not.

    Inputs:
    game_id          - String representing game id

    Outputs:
    v2_dict          - Dictionary of the JSON response from the stats.nba.com api
    """
    v2_api_url = (
        "https://stats.nba.com/stats/playbyplayv2?"
        f"EndPeriod=14&GameID={game_id}&StartPeriod=1"
    )

    try:
        v2_rep = requests.get(v2_api_url, headers=USER_AGENT)
    except json.decoder.JSONDecodeError as ex:
        print(ex)
        print(f"This is the stats.nba.com API's output: {v2_rep.text}")
        sys.exit()

    v2_dict = v2_rep.json()

    return v2_dict


def get_lineup_api(game_id, period):
    """
    function pulls the possible lineups for the given period and game id for
    both the away and home teams

    Inputs:
    game_id            - id of game
    period             - period of game

    Outputs:
    lineup_req_dict    - dictionary of lineup api request respons
    """

    if period <= 4:
        start_range = (((period - 1) * 720) * 10) + 5
        if game_id == "0020100810" and period == 4:
            end_range = start_range + 3000
        else:
            end_range = start_range + 1000
    else:
        start_range = ((((period - 5) * 300) + 2880) * 10) + 5
        end_range = start_range + 1000

    url = (
        f"https://stats.nba.com/stats/boxscoreadvancedv2/?gameId={game_id}&"
        f"startPeriod={period}&endPeriod={period}&startRange={start_range}&"
        f"endRange={end_range}&rangeType=2"
    )

    lineups_req = requests.get(url, headers=USER_AGENT)
    lineup_req_dict = json.loads(lineups_req.text)

    return lineup_req_dict


def get_lineup(period_df, lineups, dataframe):
    """
    this function calculates the lineups for each team at each event and then
    appends it to the current dataframe. This only works for one period at a
    time

    Inputs:
    period_df         - the main game pbp dataframe subsetted to only one period
                        in the game
    lineups           - lineup api response dictionary
    dataframe         - full game dataframe. This is passed to get players name from
                        id in case the player didn't have an event in that period.

    Outputs:
    lineup_df     - period_df with each teams lineups calculate and added to the
                    dataframe
    """

    # subsets main dataframe by period and subsets into a home and away subs
    subs_df = period_df[(period_df.event_type_de == "substitution")]
    away_subs = subs_df[pd.isnull(subs_df["visitordescription"]) == 0]
    home_subs = subs_df[pd.isnull(subs_df["homedescription"]) == 0]

    home_team = period_df["home_team_id"].unique()[0]
    away_team = period_df["away_team_id"].unique()[0]
    players = lineups["resultSets"][0]["rowSet"]
    home_ids_names = [(p[4], p[5]) for p in players if p[1] == home_team]
    away_ids_names = [(p[4], p[5]) for p in players if p[1] == away_team]

    # gets the index of the first sub for home and away to get the players who started
    # the period by subsetting the dataframe to all actions before the first sub for
    # each team
    away_indexes = list(away_subs.index)
    home_indexes = list(home_subs.index)
    # pulls the unique values from the whole period dataframe if there are no subs
    # then it just pulls the unique ids from the from the dataframe itself because
    # the away/home indexes will be an empty list

    try:
        home_starting_line = list(
            period_df[
                (period_df.event_team == period_df["home_team_abbrev"].unique()[0])
                & (~pd.isnull(period_df["player1_name"]))
                & (
                    period_df["player1_team_abbreviation"]
                    == period_df["home_team_abbrev"].unique()[0]
                )
                & (period_df.is_block == 0)
                & (period_df.is_steal == 0)
            ]
            .loc[: home_indexes[0], :]["player1_id"]
            .unique()
        )
    except IndexError:
        home_starting_line = list(
            period_df[
                (period_df.event_team == period_df["home_team_abbrev"].unique()[0])
                & (~pd.isnull(period_df["player1_name"]))
                & (
                    period_df["player1_team_abbreviation"]
                    == period_df["home_team_abbrev"].unique()[0]
                )
                & (period_df.is_block == 0)
                & (period_df.is_steal == 0)
            ]["player1_id"].unique()
        )

    if {x for x in home_starting_line} != {x[0] for x in home_ids_names} or len(
        home_ids_names
    ) != 5:
        starting_lineup = set()
        subs = set()
        for i in range(period_df.shape[0]):
            if (
                period_df.iloc[i, :]["event_team"]
                == period_df["home_team_abbrev"].unique()[0]
                and pd.isnull(period_df.iloc[i, :]["player1_name"]) != 1
                and period_df.iloc[i, :]["player1_team_abbreviation"]
                == period_df.iloc[i, :]["home_team_abbrev"]
                and period_df.iloc[i, :]["is_block"] == 0
                and period_df.iloc[i, :]["is_steal"] == 0
            ):
                if period_df.iloc[i, :]["event_type_de"] != "substitution":
                    if (
                        period_df.iloc[i, :]["player1_id"] != 0
                        and period_df.iloc[i, :]["player1_id"] not in subs
                    ):
                        starting_lineup.add(period_df.iloc[i, :]["player1_id"])
                else:
                    if period_df.iloc[i, :]["player2_id"] not in starting_lineup:
                        subs.add(period_df.iloc[i, :]["player2_id"])
                    if period_df.iloc[i, :]["player1_id"] not in subs:
                        starting_lineup.add(period_df.iloc[i, :]["player1_id"])

                if len(starting_lineup) == 5:
                    break
        if len(home_ids_names) < 5:
            home_ids_names = [
                (
                    period_df[period_df.player1_id == x].player1_id.unique()[0],
                    period_df[period_df.player1_id == x].player1_name.unique()[0],
                )
                for x in starting_lineup
            ]
        else:
            home_ids_names = [(p[0], p[1]) for p in home_ids_names if p[0] not in subs]
        if len(home_ids_names) != 5 and len(starting_lineup) == 5:
            home_ids_names = [
                (
                    period_df[period_df.player1_id == x].player1_id.unique()[0],
                    period_df[period_df.player1_id == x].player1_name.unique()[0],
                )
                for x in starting_lineup
            ]

    try:
        away_starting_line = list(
            period_df[
                (period_df.event_team == period_df["away_team_abbrev"].unique()[0])
                & (~pd.isnull(period_df["player1_name"]))
                & (
                    period_df["player1_team_abbreviation"]
                    == period_df["away_team_abbrev"].unique()[0]
                )
                & (period_df.is_block == 0)
                & (period_df.is_steal == 0)
            ]
            .loc[: away_indexes[0], :]["player1_id"]
            .unique()
        )
    except IndexError:
        away_starting_line = list(
            period_df[
                (period_df.event_team == period_df["away_team_abbrev"].unique()[0])
                & (~pd.isnull(period_df["player1_name"]))
                & (
                    period_df["player1_team_abbreviation"]
                    == period_df["away_team_abbrev"].unique()[0]
                )
                & (period_df.is_block == 0)
                & (period_df.is_steal == 0)
            ]["player1_id"].unique()
        )
    if {x for x in away_starting_line} != {x[0] for x in away_ids_names} or len(
        away_ids_names
    ) != 5:
        starting_lineup = set()
        subs = set()
        for i in range(period_df.shape[0]):
            if (
                period_df.iloc[i, :]["event_team"]
                == period_df["away_team_abbrev"].unique()[0]
                and pd.isnull(period_df.iloc[i, :]["player1_name"]) != 1
                and period_df.iloc[i, :]["player1_team_abbreviation"]
                == period_df.iloc[i, :]["away_team_abbrev"]
                and period_df.iloc[i, :]["is_block"] == 0
                and period_df.iloc[i, :]["is_steal"] == 0
            ):
                if period_df.iloc[i, :]["event_type_de"] != "substitution":
                    if (
                        period_df.iloc[i, :]["player1_id"] != 0
                        and period_df.iloc[i, :]["player1_id"] not in subs
                    ):
                        starting_lineup.add(period_df.iloc[i, :]["player1_id"])
                else:
                    if period_df.iloc[i, :]["player2_id"] not in starting_lineup:
                        subs.add(period_df.iloc[i, :]["player2_id"])
                    if period_df.iloc[i, :]["player1_id"] not in subs:
                        starting_lineup.add(period_df.iloc[i, :]["player1_id"])

                if len(starting_lineup) == 5:
                    break
        if len(away_ids_names) < 5:
            away_ids_names = [
                (
                    period_df[period_df.player1_id == x].player1_id.unique()[0],
                    period_df[period_df.player1_id == x].player1_name.unique()[0],
                )
                for x in starting_lineup
            ]
        else:
            away_ids_names = [(p[0], p[1]) for p in away_ids_names if p[0] not in subs]
        if len(away_ids_names) != 5 and len(starting_lineup) == 5:
            away_ids_names = [
                (
                    period_df[period_df.player1_id == x].player1_id.unique()[0],
                    period_df[period_df.player1_id == x].player1_name.unique()[0],
                )
                for x in starting_lineup
            ]

    # creating columns to populate with players on the court
    period_df.loc[:, "home_player_1"] = ""
    period_df.loc[:, "home_player_1_id"] = ""
    period_df.loc[:, "home_player_2"] = ""
    period_df.loc[:, "home_player_2_id"] = ""
    period_df.loc[:, "home_player_3"] = ""
    period_df.loc[:, "home_player_3_id"] = ""
    period_df.loc[:, "home_player_4"] = ""
    period_df.loc[:, "home_player_4_id"] = ""
    period_df.loc[:, "home_player_5"] = ""
    period_df.loc[:, "home_player_5_id"] = ""
    period_df.loc[:, "away_player_1"] = ""
    period_df.loc[:, "away_player_1_id"] = ""
    period_df.loc[:, "away_player_2"] = ""
    period_df.loc[:, "away_player_2_id"] = ""
    period_df.loc[:, "away_player_3"] = ""
    period_df.loc[:, "away_player_3_id"] = ""
    period_df.loc[:, "away_player_4"] = ""
    period_df.loc[:, "away_player_4_id"] = ""
    period_df.loc[:, "away_player_5"] = ""
    period_df.loc[:, "away_player_5_id"] = ""
    # add players to the columns by looping through the dataframe and putting the
    # players in for each row using the starting lineup list. If there is a
    # substitution event then the player coming on replaces the player going off in
    # the list this is done for the whole period
    if (
        period_df.game_id.unique()[0] == "0020200992"
        and period_df.period.unique()[0] == 5
    ):
        away_ids_names.append((922, "Elden Campbell"))

    for i in range(period_df.shape[0]):
        if (
            period_df.iloc[i, :]["event_type_de"] == "substitution"
            and pd.isnull(period_df.iloc[i, :]["visitordescription"]) == 1
        ):
            home_ids_names = [
                ids
                for ids in home_ids_names
                if ids[0] != period_df.iloc[i, :]["player1_id"]
            ]
            home_ids_names.append((period_df.iloc[i, 20], period_df.iloc[i, 21]))
            period_df.iat[i, 57] = home_ids_names[0][0]
            period_df.iat[i, 56] = home_ids_names[0][1]
            period_df.iat[i, 59] = home_ids_names[1][0]
            period_df.iat[i, 58] = home_ids_names[1][1]
            period_df.iat[i, 61] = home_ids_names[2][0]
            period_df.iat[i, 60] = home_ids_names[2][1]
            period_df.iat[i, 63] = home_ids_names[3][0]
            period_df.iat[i, 62] = home_ids_names[3][1]
            period_df.iat[i, 65] = home_ids_names[4][0]
            period_df.iat[i, 64] = home_ids_names[4][1]
            period_df.iat[i, 67] = away_ids_names[0][0]
            period_df.iat[i, 66] = away_ids_names[0][1]
            period_df.iat[i, 69] = away_ids_names[1][0]
            period_df.iat[i, 68] = away_ids_names[1][1]
            period_df.iat[i, 71] = away_ids_names[2][0]
            period_df.iat[i, 70] = away_ids_names[2][1]
            period_df.iat[i, 73] = away_ids_names[3][0]
            period_df.iat[i, 72] = away_ids_names[3][1]
            period_df.iat[i, 75] = away_ids_names[4][0]
            period_df.iat[i, 74] = away_ids_names[4][1]
        elif (
            period_df.iloc[i, :]["event_type_de"] == "substitution"
            and pd.isnull(period_df.iloc[i, :]["homedescription"]) == 1
        ):
            away_ids_names = [
                ids
                for ids in away_ids_names
                if ids[0] != period_df.iloc[i, :]["player1_id"]
            ]
            away_ids_names.append((period_df.iloc[i, 20], period_df.iloc[i, 21]))
            period_df.iat[i, 57] = home_ids_names[0][0]
            period_df.iat[i, 56] = home_ids_names[0][1]
            period_df.iat[i, 59] = home_ids_names[1][0]
            period_df.iat[i, 58] = home_ids_names[1][1]
            period_df.iat[i, 61] = home_ids_names[2][0]
            period_df.iat[i, 60] = home_ids_names[2][1]
            period_df.iat[i, 63] = home_ids_names[3][0]
            period_df.iat[i, 62] = home_ids_names[3][1]
            period_df.iat[i, 65] = home_ids_names[4][0]
            period_df.iat[i, 64] = home_ids_names[4][1]
            period_df.iat[i, 67] = away_ids_names[0][0]
            period_df.iat[i, 66] = away_ids_names[0][1]
            period_df.iat[i, 69] = away_ids_names[1][0]
            period_df.iat[i, 68] = away_ids_names[1][1]
            period_df.iat[i, 71] = away_ids_names[2][0]
            period_df.iat[i, 70] = away_ids_names[2][1]
            period_df.iat[i, 73] = away_ids_names[3][0]
            period_df.iat[i, 72] = away_ids_names[3][1]
            period_df.iat[i, 75] = away_ids_names[4][0]
            period_df.iat[i, 74] = away_ids_names[4][1]
        else:
            period_df.iat[i, 57] = home_ids_names[0][0]
            period_df.iat[i, 56] = home_ids_names[0][1]
            period_df.iat[i, 59] = home_ids_names[1][0]
            period_df.iat[i, 58] = home_ids_names[1][1]
            period_df.iat[i, 61] = home_ids_names[2][0]
            period_df.iat[i, 60] = home_ids_names[2][1]
            period_df.iat[i, 63] = home_ids_names[3][0]
            period_df.iat[i, 62] = home_ids_names[3][1]
            period_df.iat[i, 65] = home_ids_names[4][0]
            period_df.iat[i, 64] = home_ids_names[4][1]
            period_df.iat[i, 67] = away_ids_names[0][0]
            period_df.iat[i, 66] = away_ids_names[0][1]
            period_df.iat[i, 69] = away_ids_names[1][0]
            period_df.iat[i, 68] = away_ids_names[1][1]
            period_df.iat[i, 71] = away_ids_names[2][0]
            period_df.iat[i, 70] = away_ids_names[2][1]
            period_df.iat[i, 73] = away_ids_names[3][0]
            period_df.iat[i, 72] = away_ids_names[3][1]
            period_df.iat[i, 75] = away_ids_names[4][0]
            period_df.iat[i, 74] = away_ids_names[4][1]

    return period_df


def main_scrape(game_id):
    """
    this is the main function that runs and ties all them together. Doing it
    this way so I can better write tests that work on Travis CI due to their
    IP being blacklisted by NBA.com.

    Inputs:
    game_id     - NBA game id of game to be scraped

    Outputs:
    game_df     - pandas dataframe of the play by play
    """

    v2_dict = get_pbp_api(game_id)
    game_df = scrape_pbp(v2_dict)
    periods = []
    if game_id == "0021500916":
        game_df = game_df[game_df["period"] < 5]
    for period in range(1, game_df["period"].max() + 1):
        lineups = get_lineup_api(game_id, period)
        periods.append(
            get_lineup(game_df[game_df["period"] == period].copy(), lineups, game_df,)
        )
    game_df = pd.concat(periods).reset_index(drop=True)
    season_dict = {
        "1": "Pre+Season",
        "2": "Regular+Season",
        "3": "All+Star",
        "4": "Playoffs",
    }
    season_type = season_dict[game_df["game_id"].unique()[0][2:3]]
    if game_df["game_id"].unique()[0][3:5] == "99":
        season = "1999-00"
    else:
        season = f"20{game_df['game_id'].unique()[0][3:5]}-{int(game_df['game_id'].unique()[0][3:5]) + 1}"
    date_url = (
        f"https://stats.nba.com/stats/teamgamelog?DateFrom=&DateTo=&LeagueID=&"
        f"Season={season}"
        f"&SeasonType={season_type}&TeamID={game_df['home_team_id'].unique()[0]}"
    )
    dates = requests.get(date_url, headers=USER_AGENT)
    dates_dict = json.loads(dates.text)
    schedule = dates_dict["resultSets"][0]["rowSet"]
    game_date = [g[2] for g in schedule if g[1] == game_df["game_id"].unique()[0]]
    formatted_date = datetime.datetime.strptime(game_date[0], "%b %d, %Y")
    game_df["game_date"] = formatted_date

    return game_df



"""
These are the functions to calculate various statistics or to pull out/parse certain data types from the play by play. They have nothing to do with the functionality of actually scraping the
data.
"""
import numpy as np


foul_dict = {
    0: "no foul",
    1: "personal",
    2: "shooting",
    3: "loose ball",
    4: "offensive",
    5: "inbound",
    6: "away from play",
    7: "elbow",
    8: "punching",
    9: "clear path",
    10: "double personal",
    11: "technical",
    12: "non-unsportsmanlike technical",
    13: "hanging technical",
    14: "flagrant type 1",
    15: "flagrant type 2",
    16: "double technical",
    17: "defense 3 second",
    18: "delay technical",
    19: "taunting technical",
    20: "player_control",
    21: "indirect technical",
    25: "excess timeout technical",
    26: "offensive charge",
    27: "personal block",
    28: "shooting block",
    30: "too many players technical",
}
SHOT_DICT = {
    1: {
        0: "No Shot",
        1: "Jump Shot",
        2: "Running Jump Shot",
        3: "Hook Shot",
        4: "Tip Shot",
        5: "Layup Shot",
        6: "Driving Layup Shot",
        7: "Dunk Shot",
        8: "Slam Dunk Shot",
        9: "Driving Dunk Shot",
        40: "Layup Shot",
        41: "Running Layup Shot",
        42: "Driving Layup Shot",
        43: "Alley Oop Layup shot",
        44: "Reverse Layup Shot",
        45: "Jump Shot",
        46: "Running Jump Shot",
        47: "Turnaround Jump Shot",
        48: "Dunk Shot",
        49: "Driving Dunk Shot",
        50: "Running Dunk Shot",
        51: "Reverse Dunk Shot",
        52: "Alley Oop Dunk Shot",
        53: "Tip Shot",
        54: "Running Tip Shot",
        55: "Hook Shot",
        56: "Running Hook Shot",
        57: "Driving Hook Shot",
        58: "Turnaround Hook Shot",
        59: "Finger Roll Shot",
        60: "Running Finger Roll Shot",
        61: "Driving Finger Roll Shot",
        62: "Turnaround Finger Roll Shot",
        63: "Fadeaway Jump Shot",
        64: "Follow Up Dunk Shot",
        65: "Jump Hook Shot",
        66: "Jump Bank Shot",
        67: "Hook Bank Shot",
        71: "Finger Roll Layup Shot",
        72: "Putback Layup Shot",
        73: "Driving Reverse Layup Shot",
        74: "Running Reverse Layup Shot",
        75: "Driving Finger Roll Layup Shot",
        76: "Running Finger Roll Layup Shot",
        77: "Driving Jump shot",
        78: "Floating Jump shot",
        79: "Pullup Jump shot",
        80: "Step Back Jump shot",
        81: "Pullup Bank shot",
        82: "Driving Bank shot",
        83: "Fadeaway Bank shot",
        84: "Running Bank shot",
        85: "Turnaround Bank shot",
        86: "Turnaround Fadeaway shot",
        87: "Putback Dunk Shot",
        88: "Driving Slam Dunk Shot",
        89: "Reverse Slam Dunk Shot",
        90: "Running Slam Dunk Shot",
        91: "Putback Reverse Dunk Shot",
        92: "Putback Slam Dunk Shot",
        93: "Driving Bank Hook Shot",
        94: "Jump Bank Hook Shot",
        95: "Running Bank Hook Shot",
        96: "Turnaround Bank Hook Shot",
        97: "Tip Layup Shot",
        98: "Cutting Layup Shot",
        99: "Cutting Finger Roll Layup Shot",
        100: "Running Alley Oop Layup Shot",
        101: "Driving Floating Jump Shot",
        102: "Driving Floating Bank Jump Shot",
        103: "Running Pull-Up Jump Shot",
        104: "Step Back Bank Jump Shot",
        105: "Turnaround Fadeaway Bank Jump Shot",
        106: "Running Alley Oop Dunk Shot",
        107: "Tip Dunk Shot",
        108: "Cutting Dunk Shot",
        109: "Driving Reverse Dunk Shot",
        110: "Running Reverse Dunk Shot",
    },
    2: {
        0: "No Shot",
        1: "Jump Shot",
        2: "Running Jump Shot",
        3: "Hook Shot",
        4: "Tip Shot",
        5: "Layup Shot",
        6: "Driving Layup Shot",
        7: "Dunk Shot",
        8: "Slam Dunk Shot",
        9: "Driving Dunk Shot",
        40: "Layup Shot",
        41: "Running Layup Shot",
        42: "Driving Layup Shot",
        43: "Alley Oop Layup shot",
        44: "Reverse Layup Shot",
        45: "Jump Shot",
        46: "Running Jump Shot",
        47: "Turnaround Jump Shot",
        48: "Dunk Shot",
        49: "Driving Dunk Shot",
        50: "Running Dunk Shot",
        51: "Reverse Dunk Shot",
        52: "Alley Oop Dunk Shot",
        53: "Tip Shot",
        54: "Running Tip Shot",
        55: "Hook Shot",
        56: "Running Hook Shot",
        57: "Driving Hook Shot",
        58: "Turnaround Hook Shot",
        59: "Finger Roll Shot",
        60: "Running Finger Roll Shot",
        61: "Driving Finger Roll Shot",
        62: "Turnaround Finger Roll Shot",
        63: "Fadeaway Jump Shot",
        64: "Follow Up Dunk Shot",
        65: "Jump Hook Shot",
        66: "Jump Bank Shot",
        67: "Hook Bank Shot",
        71: "Finger Roll Layup Shot",
        72: "Putback Layup Shot",
        73: "Driving Reverse Layup Shot",
        74: "Running Reverse Layup Shot",
        75: "Driving Finger Roll Layup Shot",
        76: "Running Finger Roll Layup Shot",
        77: "Driving Jump shot",
        78: "Floating Jump shot",
        79: "Pullup Jump shot",
        80: "Step Back Jump shot",
        81: "Pullup Bank shot",
        82: "Driving Bank shot",
        83: "Fadeaway Bank shot",
        84: "Running Bank shot",
        85: "Turnaround Bank shot",
        86: "Turnaround Fadeaway shot",
        87: "Putback Dunk Shot",
        88: "Driving Slam Dunk Shot",
        89: "Reverse Slam Dunk Shot",
        90: "Running Slam Dunk Shot",
        91: "Putback Reverse Dunk Shot",
        92: "Putback Slam Dunk Shot",
        93: "Driving Bank Hook Shot",
        94: "Jump Bank Hook Shot",
        95: "Running Bank Hook Shot",
        96: "Turnaround Bank Hook Shot",
        97: "Tip Layup Shot",
        98: "Cutting Layup Shot",
        99: "Cutting Finger Roll Layup Shot",
        100: "Running Alley Oop Layup Shot",
        101: "Driving Floating Jump Shot",
        102: "Driving Floating Bank Jump Shot",
        103: "Running Pull-Up Jump Shot",
        104: "Step Back Bank Jump Shot",
        105: "Turnaround Fadeaway Bank Jump Shot",
        106: "Running Alley Oop Dunk Shot",
        107: "Tip Dunk Shot",
        108: "Cutting Dunk Shot",
        109: "Driving Reverse Dunk Shot",
        110: "Running Reverse Dunk Shot",
    },
    3: {
        0: "No Shot",
        10: "Free Throw 1 of 1",
        11: "Free Throw 1 of 2",
        12: "Free Throw 2 of 2",
        13: "Free Throw 1 of 3",
        14: "Free Throw 2 of 3",
        15: "Free Throw 3 of 3",
        16: "Free Throw Technical",
        17: "Free Throw Clear Path",
        18: "Free Throw Flagrant 1 of 2",
        19: "Free Throw Flagrant 2 of 2",
        20: "Free Throw Flagrant 1 of 1",
        21: "Free Throw Technical 1 of 2",
        22: "Free Throw Technical 2 of 2",
        25: "Free Throw Clear Path 1 of 2",
        26: "Free Throw Clear Path 2 of 2",
        27: "Free Throw Flagrant 1 of 3",
        28: "Free Throw Flagrant 2 of 3",
        29: "Free Throw Flagrant 3 of 3",
    },
}


def wnba_made_shot(row):
    """
    function to determine whether shot was made or missed

    Input:
    row - pandas row

    Output:
    shot_made - binary variable
    """

    de = row["de"]

    if row["etype"] == 1:
        return 1
    elif row["etype"] == 2:
        return 0
    elif row["etype"] == 3 and ("Missed" in de):
        return 0
    elif row["etype"] == 3:
        return 1
    else:
        return np.nan


def made_shot(row):
    """
    function to determine whether shot was made or missed

    Input:
    row - pandas row

    Output:
    shot_made - binary variable
    """

    if row["homedescription"] == None:
        home_d = "false"
    else:
        home_d = row["homedescription"]

    if row["visitordescription"] == None:
        away_d = "false"
    else:
        away_d = row["visitordescription"]
    if row["eventmsgtype"] == 1:
        return 1
    elif row["eventmsgtype"] == 2:
        return 0
    elif row["eventmsgtype"] == 3 and ("MISS" in home_d or "MISS" in away_d):
        return 0
    elif row["eventmsgtype"] == 3:
        return 1
    else:
        return np.nan


def wnba_parse_foul(row):
    """
    function to determine what type of foul is being commited by the player

    Input:
    row - row of nba play by play

    Output:
    foul_type - the foul type of the fould commited by the player
    """

    try:
        if row["etype"] == 6:
            try:
                return foul_dict[row["mtype"]]
            except KeyError:
                return np.nan
        return np.nan
    except KeyError:
        return np.nan


def parse_foul(row):
    """
    function to determine what type of foul is being commited by the player

    Input:
    row - row of nba play by play

    Output:
    foul_type - the foul type of the fould commited by the player
    """

    try:
        if row["eventmsgtype"] == 6:
            try:
                return foul_dict[row["eventmsgactiontype"]]
            except KeyError:
                return np.nan
        return np.nan
    except KeyError:
        return np.nan


def wnba_shot_types(row):
    """
    function to parse what type of shot is being taken

    Inputs:
    row - pandas row of play by play dataframe

    Outputs:
    shot_type - returns a shot type of the values hook, jump, layup, dunk, tip
    """
    try:
        if row["etype"] in [1, 2, 3]:
            return SHOT_DICT[row["etype"]][row["mtype"]]
        else:
            return np.nan
    except KeyError:
        return np.nan


def parse_shot_types(row):
    """
    function to parse what type of shot is being taken

    Inputs:
    row - pandas row of play by play dataframe

    Outputs:
    shot_type - returns a shot type of the values hook, jump, layup, dunk, tip
    """
    try:
        if row["eventmsgtype"] in [1, 2, 3]:
            return SHOT_DICT[row["eventmsgtype"]][row["eventmsgactiontype"]]
        else:
            return np.nan
    except KeyError:
        return np.nan


def wnba_seconds_elapsed(row):
    """
    this function parses the string time column and converts it into game
    seconds elapsed

    Inputs:
    row - row of play by play dataframe

    Outputs:
    time_in_seconds - the elapsed game time expressed in seconds
    """

    time_list = row["cl"].strip().split(":")
    max_time = 600
    ot_max_time = 300

    if row["period"] in [1, 2, 3, 4]:
        time_in_seconds = (
            max_time - (int(time_list[0]) * 60 + float(time_list[1]))
        ) + (720 * (int(row["period"]) - 1))
    elif row["period"] > 4:
        time_in_seconds = (
            (ot_max_time - (int(time_list[0]) * 60 + float(time_list[1])))
            + (300 * (int(row["period"]) - 5))
            + 2880
        )

    return time_in_seconds


def create_seconds_elapsed(row):
    """
    this function parses the string time column and converts it into game
    seconds elapsed

    Inputs:
    row - row of play by play dataframe

    Outputs:
    time_in_seconds - the elapsed game time expressed in seconds
    """

    time_list = row["pctimestring"].strip().split(":")
    max_time = 720
    ot_max_time = 300

    if row["period"] in [1, 2, 3, 4]:
        time_in_seconds = (max_time - (int(time_list[0]) * 60 + int(time_list[1]))) + (
            720 * (int(row["period"]) - 1)
        )
    elif row["period"] > 4:
        time_in_seconds = (
            (ot_max_time - (int(time_list[0]) * 60 + int(time_list[1])))
            + (300 * (int(row["period"]) - 5))
            + 2880
        )

    return time_in_seconds


def wnba_points_made(row):
    """
    function to calculate the points earned by a team with each shot made

    Inputs:
    row - row of pbp dataframe

    Outputs - value of shot made
    """

    if row["is_three"] == 1 and row["shot_made"] == 1:
        return 3
    elif row["is_three"] == 0 and row["shot_made"] == 1 and row["etype"] != 3:
        return 2
    elif row["etype"] == 3 and row["shot_made"] == 1:
        return 1
    else:
        return 0


def calc_points_made(row):
    """
    function to calculate the points earned by a team with each shot made

    Inputs:
    row - row of pbp dataframe

    Outputs - value of shot made
    """

    if row["is_three"] == 1 and row["shot_made"] == 1:
        return 3
    elif row["is_three"] == 0 and row["shot_made"] == 1 and row["eventmsgtype"] != 3:
        return 2
    elif row["eventmsgtype"] == 3 and row["shot_made"] == 1:
        return 1
    else:
        return 0
